﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using Henge3D.Physics;

namespace Henge3D.Pipeline
{
    public enum MeshType
    {
        Both,
        Physical,
        Visual
    }

    public enum PhysicalShape
    {
        Mesh,
        Polyhedron,
        Sphere,
        Capsule,
        Cube
    }

    public enum WindingOrder
    {
        Clockwise,
        CounterClockwise
    }

    public class RigidBodyModel
	{
		[ContentSerializer]
		private MassProperties _mass;
		[ContentSerializer]
		private CompiledPart[] _parts;
		[ContentSerializer]
		private Material[] _materials;

		public MassProperties MassProperties { get { return _mass; } }
		public CompiledPart[] Parts { get { return _parts; } }
		public Material[] Materials { get { return _materials; } }

        private float _defaultDensity = 1.0f, _defaultElasticity = 0f, _defaultRoughness = 0.5f;

        private RigidBodyModel()
		{
		}

        public RigidBodyModel(MassProperties mass, CompiledPart[] parts, Material[] materials)
        {
            Setup(mass, parts, materials);
        }
        private void Setup(MassProperties mass, CompiledPart[] parts, Material[] materials)
        { 
			_mass = mass;
			_parts = parts;
			_materials = materials;

			if (_parts.Length != _materials.Length)
			{
				throw new ArgumentException("The count of supplied mesh parts and materials do not match.");
			}
		}

        public RigidBodyModel(Model model, float? elasticity, float? roughness, float? density, PhysicalShape shape)
        {
            var parts = new List<CompiledPart>();
            var materials = new List<Material>();
            var mass = new MassProperties();
            var centerOfMass = Vector3.Zero;

            if (!elasticity.HasValue)
                elasticity = _defaultElasticity;
            if (!roughness.HasValue)
                roughness = _defaultRoughness;
            if (!density.HasValue)
                density = _defaultDensity;

            foreach (var mesh in model.Meshes)
            {
                var meshCenterOfMass = Vector3.Zero;
                var meshMass = MassProperties.Immovable;
                CompiledPart meshPart = null;

                if (mesh.MeshParts.Count < 1)
                {
                    continue;
                }

                int[] indices = new int[mesh.MeshParts[0].PrimitiveCount * 3];
                ushort[] ri = new ushort[mesh.MeshParts[0].PrimitiveCount * 3];
                mesh.MeshParts[0].IndexBuffer.GetData<ushort>(ri, mesh.MeshParts[0].StartIndex, mesh.MeshParts[0].PrimitiveCount * 3);
                for (int i = 0; i < ri.Length; i++)
                    indices[i] = (int)ri[i];

                Vector3[] vertices = MeshToVertexArray(mesh);


                switch (shape)
                {
                    case PhysicalShape.Cube:
                        {
                            meshPart = new CompiledMesh(vertices, indices);
                            meshCenterOfMass = GetMeshTranslation(mesh);
                            Vector3 size = new Vector3(0, 0, 0);
                            for (int i=0; i<vertices.Length; i++)
                            {
                                if (Math.Abs(vertices[i].X) > size.X)
                                    size.X = Math.Abs(vertices[i].X);
                                if (Math.Abs(vertices[i].Y) > size.Y)
                                    size.Y = Math.Abs(vertices[i].Y);
                                if (Math.Abs(vertices[i].Z) > size.Z)
                                    size.Z = Math.Abs(vertices[i].Z);
                            }
                            meshMass = MassProperties.FromCuboid(density.Value, size);
                        }
                        break;
                    case PhysicalShape.Mesh:
                        {
                            meshPart = new CompiledMesh(vertices, indices);
                            meshMass = MassProperties.Immovable;
                            meshCenterOfMass = GetMeshTranslation(mesh);
                        }
                        break;
                    case PhysicalShape.Polyhedron:
                        {
                            var hull = new ConvexHull3D(vertices);
                            meshPart = hull.ToPolyhedron();
                            meshMass = MassProperties.FromTriMesh(density.Value, vertices, indices, out meshCenterOfMass);
                        }
                        break;
                    case PhysicalShape.Sphere:
                        {
                            Sphere s;
                            Sphere.Fit(vertices, out s);
                            meshPart = new CompiledSphere(s.Center, s.Radius);
                            meshMass = MassProperties.FromSphere(density.Value, s.Center, s.Radius);
                            meshCenterOfMass = s.Center;
                        }
                        break;
                    case PhysicalShape.Capsule:
                        {
                            Capsule c;
                            Capsule.Fit(vertices, out c);
                            meshPart = new CompiledCapsule(c.P1, c.P2, c.Radius);
                            meshMass = MassProperties.FromCapsule(density.Value, c.P1, c.P2, c.Radius, out meshCenterOfMass);
                        }
                        break;
                }
                parts.Add(meshPart);
                materials.Add(new Material(elasticity.Value, roughness.Value));
                Vector3.Multiply(ref meshCenterOfMass, meshMass.Mass, out meshCenterOfMass);
                Vector3.Add(ref centerOfMass, ref meshCenterOfMass, out centerOfMass);
                mass.Mass += meshMass.Mass;
                meshMass.Inertia.M44 = 0f;
                Matrix.Add(ref mass.Inertia, ref meshMass.Inertia, out mass.Inertia);


                // compute mass properties
                Vector3.Divide(ref centerOfMass, mass.Mass, out centerOfMass);
                mass.Inertia.M44 = 1f;
                MassProperties.TranslateInertiaTensor(ref mass.Inertia, -mass.Mass, centerOfMass, out mass.Inertia);
                if (centerOfMass.Length() >= Constants.Epsilon)
                {
                    var transform = Matrix.CreateTranslation(-centerOfMass.X, -centerOfMass.Y, -centerOfMass.Z);
                    foreach (var p in parts)
                    {
                        p.Transform(ref transform);
                    }

                    transform = model.Root.Transform;
                    transform.M41 -= centerOfMass.X;
                    transform.M42 -= centerOfMass.Y;
                    transform.M43 -= centerOfMass.Z;
                    model.Root.Transform = transform;
                }

                mass = new MassProperties(mass.Mass, mass.Inertia);
                Setup(mass, parts.ToArray(), materials.ToArray());
            }
        }


        private Vector3[] MeshToVertexArray(ModelMesh mesh)
        {
            MemoryStream ms;
            var buffer = mesh.MeshParts[0].VertexBuffer;

            byte[] bb = new byte[buffer.VertexCount * buffer.VertexDeclaration.VertexStride];
            buffer.GetData<byte>(bb);

            ms = new MemoryStream(bb);

            BinaryReader reader = new BinaryReader(ms);

            var elems = buffer.VertexDeclaration.GetVertexElements();
            int count = mesh.MeshParts[0].NumVertices;

            ms.Seek(mesh.MeshParts[0].VertexOffset * buffer.VertexDeclaration.VertexStride, SeekOrigin.Begin);

            var vertices = new Vector3[count];
            for (int i = 0; i < count; i++)
            {
                foreach (var elType in elems)
                {
                    if (elType.VertexElementUsage == VertexElementUsage.Position)
                    {
                        vertices[i].X = reader.ReadSingle();
                        vertices[i].Y = reader.ReadSingle();
                        vertices[i].Z = reader.ReadSingle();
                    }
                    else
                    {
                        switch (elType.VertexElementFormat)
                        {
                            case VertexElementFormat.Color:
                                reader.ReadUInt32();
                                break;
                            case VertexElementFormat.Vector2:
                                reader.ReadSingle();
                                reader.ReadSingle();
                                break;
                            case VertexElementFormat.Vector3:
                                reader.ReadSingle();
                                reader.ReadSingle();
                                reader.ReadSingle();
                                break;
                            default:
                                throw new Exception("Unrecognized element type in vertex buffer: " + elType.ToString());
                        }
                    }
                }
            }

            var transforms = new Stack<Matrix>();
            var bone = mesh.ParentBone;
            while (bone != null)
            {
                transforms.Push(bone.Transform);
                bone = bone.Parent;
            }

            var transform = Matrix.Identity;
            while (transforms.Count > 0)
            {
                transform *= transforms.Pop();
            }

            Vector3.Transform(vertices, ref transform, vertices);
            return vertices;
        }

        private Vector3 GetMeshTranslation(ModelMesh mesh)
        {
            var pos = Vector3.Zero;
            var bone = mesh.ParentBone;
            while (bone != null)
            {
                pos += bone.Transform.Translation;
                bone = bone.Parent;
            }
            return pos;
        }
    }
}
